/*
Author: Jelle Geerts

Usage of the works is permitted provided that this instrument is
retained with the works, so that any entity that uses the works is
notified of this instrument.

DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.
*/

function KeePassHelper(keePassHelper_PrefMan) {
    var cfg_titleSeparatorChanged,
        cfg_urlModeChanged,
        updateTitlebar;

    KeePassHelper.separator = keePassHelper_PrefMan.getValue("separator");
    KeePassHelper.urlMode = keePassHelper_PrefMan.getUrlMode();

    this.destroy = function () {
        keePassHelper_PrefMan.removeObserver(cfg_titleSeparatorChanged);
        keePassHelper_PrefMan.removeObserver(cfg_urlModeChanged);
    };

    KeePassHelper.getContainerXulElement = function () {
        if (KeePassHelper_AppInfo.isThunderbird()) {
            return document.getElementById("tabmail");
        } else {
            return document.getElementById("content");
        }
    };

    KeePassHelper.getDocumentTitle = function (xulElement) {
        var title = "";

        if (xulElement.contentDocument) {
            title = xulElement.contentDocument.title;
        }

        if (!title && xulElement.ownerDocument) {
            title = xulElement.ownerDocument.documentElement.getAttribute("titledefault");
        }

        return title;
    };

    KeePassHelper.getHostnameFromHtmlDocument = function (doc) {
        var hostname = "",
            el,
            matches;

        try {
            hostname = doc.location.hostname;
        } catch (ex) {
            // Fails on at least 'about:config' on FX 1.5 to 3.x.
            return "";
        }

        // Workaround for the IE Tab, IE Tab 2, and Coral IE Tab add-ons.
        if ((hostname === "ietab" || hostname === "coralietab") &&
                (el = document.getElementById("security-button")) !== null) {
            hostname = el.label;
        } else if (hostname === "ietab2" && (el = doc.location.search) !== null) {
            matches = el.match(/(\?|&)url=[a-zA-Z0-9.+-]+:\/\/([-.a-zA-Z0-9]+)/);
            if (matches !== null && typeof matches[2] !== "undefined") {
                hostname = matches[2];
            }
        }

        return hostname;
    };

    KeePassHelper.getLocationFromHtmlDocument = function (doc) {
        try {
            return doc.location;
        } catch (ex) {
            dump("[KeePass Helper] getLocationFromHtmlDocument(): accessing `doc.location' resulted in exception `" + ex + "'.\n");
            return "";
        }
    };

    KeePassHelper.updateDocumentTitle = function (xulElement, originalTitle) {
        var urlMode = keePassHelper_PrefMan.getUrlMode(),
            docElement = xulElement.ownerDocument.documentElement,
            codename = docElement.getAttribute("titlemodifier"),
            location = "",
            hostname = KeePassHelper.getHostnameFromHtmlDocument(window.content.document),
            titlePrefix = "",
            matches,
            uriScheme;

        if (originalTitle.length !== 0) {
            titlePrefix = originalTitle + " " + KeePassHelper.separator + " ";
        }

        if (urlMode === "full") {
            location = KeePassHelper.getLocationFromHtmlDocument(window.content.document);
            if (location.length !== 0) {
                document.title = titlePrefix + location + " " + KeePassHelper.separator + " " + codename;
                return;
            }
        } else if (urlMode === "short") {
            location = KeePassHelper.getLocationFromHtmlDocument(window.content.document);
            matches = String(location).match(/[^:]+:(\/\/)?[^\/]*/);
            if (matches !== null && typeof(matches[0] !== "undefined")) {
                uriScheme = getUriScheme(matches[0]);
                document.title =
                    titlePrefix +
                    matches[0] +
                    // Add a trailing slash as a security measure, otherwise one is vulnerable to
                    // letting a password manager such as KeePass automatically insert credentials
                    // where they shouldn't.
                    // Example:
                    // One thinks one is visiting one's webmail interface, gmail.com, but instead,
                    // somehow, one is actually visiting gmail.com.fakesite.com. If we don't add a
                    // trailing slash, and one lets KeePass match the "gmail.com" part to recognize
                    // the window to automatically insert credentials into, then one is vulnerable
                    // to this problem. If we add a trailing slash, one can let KeePass match
                    // "gmail.com/", and thus "gmail.com.fakesite.com/" would never match that.
                    // Note that we don't include the trailing slash for locations that are
                    // built into the browser, such as "about:addons", and only for locations that
                    // have an URI scheme.
                    (uriScheme !== null ? "/" : "") +
                    " " +
                    KeePassHelper.separator +
                    " " +
                    codename;
                return;
            }
        }

        // If all else fails, or if urlMode === "hostname", then we execute this code.
        if (hostname.length === 0) {
            document.title = titlePrefix + codename;
        } else {
            document.title = titlePrefix + hostname + " " + KeePassHelper.separator + " " + codename;
        }
    };

    KeePassHelper.setAuthDialogTitle = function () {
        var extractEmailAddress = false,
            info,
            matches,
            hostname,
            account; // For Thunderbird only.

        if (!(info = document.getElementById("info.body"))) { // FX 3.0 to FX 4.x, TB 3.x
            if (!(info = document.getElementById("info.box"))) { // FX 1.5 to FX 2.x
                return;
            }
            if (!(info = info.getElementsByTagName("description")[0])) {
                return;
            }
        }

        extractEmailAddress = KeePassHelper_AppInfo.isThunderbirdOrSeaMonkey() &&
            // Username field must not be visible.
            !KeePassHelper.isElementVisible(document.getElementById("loginContainer"))

        if (extractEmailAddress) {
            /* Match an email address. Supports the quoted-string form (e.g.,
             * "John Doe"@example.org). */
            matches = info.firstChild.nodeValue.match(/(?:(?:"[a-zA-Z0-9.!#$%&'*+-\/=?^_`{|}~ ()[\]\\;:,<>]+")|(?:[a-zA-Z0-9.!#$%&'*+-\/=?^_`{|}~]+))@[a-zA-Z0-9-.@]+[a-zA-Z0-9]+/);
            if (matches !== null && typeof(matches[0] !== "undefined")) {
                account = matches[0];
            }

            if (typeof account === "undefined") {
                return;
            }
        } else {
            /* Method for:
             *   - HTTP authentication dialogs in FX 1.5 to 1.5.0.1.
             *   - HTTP NTLM authentication dialogs in FX ??? to 4.x.
             *     (For some localizations, HTTP NTLM authentication dialogs have a different
             *     message than other HTTP authentication dialogs.)
             *   - HTTP authentication dialogs in FX 1.5.0.2 to 4.x.
             */
            matches = info.firstChild.nodeValue.match(/(?:([a-zA-Z0-9.+-]+):\/\/)?((?:[-a-zA-Z0-9]+\.)+[-a-zA-Z0-9]+)/);
            if (matches === null || typeof matches[2] === "undefined") {
                // Match URLs which only specify a top-level domain, like `http://intranet'. These
                // types of URLs must have a URI scheme (`http://' here), otherwise there is no
                // "anchor", causing the regex to go haywire.
                matches = info.firstChild.nodeValue.match(/([a-zA-Z0-9.+-]+):\/\/([-a-zA-Z0-9]+)/);
            }
            if (matches !== null && typeof matches[2] !== "undefined") {
                hostname = matches[2];
            }

            if (typeof hostname === "undefined") {
                return;
            }
        }

        // Store the original title, so we can reuse it when this function is called again on the
        // same document.
        if (typeof document.keePassHelper_originalTitle === "undefined") {
            document.keePassHelper_originalTitle = document.title;
        }

        if (extractEmailAddress) {
            document.title = document.keePassHelper_originalTitle + " " +
                KeePassHelper.separator + " " + account;
        } else {
            document.title = document.keePassHelper_originalTitle + " " +
                KeePassHelper.separator + " " + hostname;
        }
    };

    // For Thunderbird only.
    this.setSetDocumentTitle = function () {
        var xulElement;

        // We don't need to do anything for SeaMonkey, as SeaMonkey's mail client does not have
        // browsing tabs. Instead, SeaMonkey's mail client opens links in the SeaMonkey browser.
        if (KeePassHelper_AppInfo.isSeaMonkey()) {
            dump("[KeePass Helper] setSetDocumentTitle(): not needed for SeaMonkey.\n");
            return;
        }

        if (!KeePassHelper_AppInfo.isThunderbird()) {
            dump("[KeePass Helper] setSetDocumentTitle(): called for a non-Thunderbird application.\n");
            return;
        }

        xulElement = KeePassHelper.getContainerXulElement();
        if (xulElement === null) {
            /* Occurs on TB 1.5 to TB 2.x. Still display a warning however, as it shouldn't occur
             * on newer Thunderbirds. */
            dump("[KeePass Helper] setSetDocumentTitle(): couldn't retrieve XULElement\n");
            return;
        }

        if (typeof(xulElement.setDocumentTitle_superFunction) !== "undefined") {
            dump("[KeePass Helper] setSetDocumentTitle(): called twice for the same XULElement.\n");
            return;
        }

        // The super function that we override might take more parameters at some point in the
        // future. It is outside our control. But to prevent incompatibilities, we can prepare for
        // change, by simply taking more arguments and passing them on to the super function, even
        // though we do not yet know in advance when they will be used and what they mean.
        xulElement.setDocumentTitle_superFunction = xulElement.setDocumentTitle;
        xulElement.setDocumentTitle = function (aTab, a, b, c, d, e) {
            // Call the super function for all tabs other than the tabs that we handle, which are
            // tabs used for browsing.
            if (typeof(aTab.browser) === "undefined") {
                this.setDocumentTitle_superFunction(aTab, a, b, c, d, e);
            } else {
                KeePassHelper.updateDocumentTitle(this, aTab.title);
            }
        };
    };

    // For Firefox/SeaMonkey only.
    this.setUpdateTitlebar = function () {
        if (!KeePassHelper_AppInfo.isFirefoxOrSeaMonkey()) {
            dump("[KeePass Helper] setUpdateTitlebar(): did _not_ detect Firefox/SeaMonkey.\n");
            return;
        }

        KeePassHelper.getContainerXulElement().updateTitlebar = function () {
            KeePassHelper.updateDocumentTitle(this, KeePassHelper.getDocumentTitle(this));
        };
    };

    cfg_titleSeparatorChanged = function (value) {
        KeePassHelper.separator = value;
        updateTitlebar();
    };

    cfg_urlModeChanged = function (value) {
        KeePassHelper.urlMode = value;
        updateTitlebar();
    };

    getUriScheme = function (s) {
        var matches = String(s).match(/([a-zA-Z0-9.+-]+):\/\//);
        if (matches !== null &&
            matches.length > 1 &&
            typeof(matches[1]) !== "undefined")
        {
            return matches[1];
        }
        return null;
    };

    updateTitlebar = function () {
        var xulElement;

        xulElement = KeePassHelper.getContainerXulElement();
        if (!xulElement) {
            KeePassHelper.setAuthDialogTitle();
            return;
        }

        if (KeePassHelper_AppInfo.isThunderbird()) {
            if (!xulElement.currentTabInfo) {
                dump("[KeePass Helper] updateTitlebar(): couldn't retrieve current tab info.\n");
            } else {
                xulElement.setDocumentTitle(xulElement.currentTabInfo);
            }
        } else {
            xulElement.updateTitlebar();
        }
    };

    keePassHelper_PrefMan.addObserver("separator", cfg_titleSeparatorChanged);
    keePassHelper_PrefMan.addObserver("url_mode", cfg_urlModeChanged);
}

KeePassHelper.isAuthDialog = function () {
    var infoBoxOK = false,
        infoContainerOK = false,
        loginContainerOK = false,
        password1ContainerOK = false;

    infoBoxOK = KeePassHelper.isElementVisible(
            document.getElementById("info.box")); // Thunderbird only. TB 1.5 to TB 2.x.
    infoContainerOK = KeePassHelper.isElementVisible(
            document.getElementById("infoContainer")); // Thunderbird only. TB 3.x.
    loginContainerOK = KeePassHelper.isElementVisible(
            document.getElementById("loginContainer")); // Firefox/SeaMonkey only.
    password1ContainerOK = KeePassHelper.isElementVisible(
            document.getElementById("password1Container"));

    return (infoBoxOK || infoContainerOK || loginContainerOK) && password1ContainerOK;
};

KeePassHelper.isElementVisible = function (el) {
    return el && !el.hidden;
};
